/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File primitives.c : User interface for SimCoupe : Drawing primitive functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"

#include "dialog.h"

/* Drawing Macros */

#define BytePos(x,y)	(((uns16)(y) << 7) + ((x) >> 1))
#define RightNybble(x)	((x) & 0x01)

#define R_MASK		0x0f
#define L_MASK		0xf0

#define FontOffset(c)	((c-32)*(GLYPH_HEIGHT*(GLYPH_WIDTH >> 1)))


#define IsOnScreen(s)	((s+1) && ((s) < 0x6000))

/* Dialog primitives */


static inline void
PlotLeft(uns16 Addr, uns8 Colour, int DrawMode)
{
	uns8 OldByte = read_byte(Addr);
	
	if (IsOnScreen(Addr))
	    switch (DrawMode) {
	
		case DM_OVER:
			write_byte(Addr,(OldByte & R_MASK) | ((Colour & R_MASK) <<4));
			break;
		case DM_AND:
			write_byte(Addr,(OldByte & R_MASK) | ((OldByte & L_MASK) & ((Colour & R_MASK) <<4)));
			break;
		case DM_OR:
			write_byte(Addr,(OldByte & R_MASK) | ((OldByte & L_MASK) | ((Colour & R_MASK) <<4)));
			break;			
		case DM_XOR:
			write_byte(Addr,(OldByte & R_MASK) | ((OldByte & L_MASK) ^ ((Colour & R_MASK) <<4)));
			break;
	    }

}
static inline void
PlotRight(uns16 Addr, uns8 Colour, int DrawMode)
{
	uns8 OldByte = read_byte(Addr);
	
	if (IsOnScreen(Addr))
	    switch (DrawMode) {
	
		case DM_OVER:
			write_byte(Addr,(OldByte & L_MASK) | (Colour & R_MASK));
			break;
		case DM_AND:
			write_byte(Addr,(OldByte & L_MASK) | ((OldByte & R_MASK) & (Colour & R_MASK)));
			break;
		case DM_OR:
			write_byte(Addr,(OldByte & L_MASK) | ((OldByte & R_MASK) | (Colour & R_MASK)));
			break;			
		case DM_XOR:
			write_byte(Addr,(OldByte & L_MASK) | ((OldByte & R_MASK) ^ (Colour & R_MASK)));
			break;
	    }

}

static inline void
PlotBoth(uns16 Addr, uns8 Colours, int DrawMode)
{
	uns8 OldByte = read_byte(Addr);
	uns8 NewByte = Colours;

	if (IsOnScreen(Addr))
	    switch (DrawMode) {
	
		case DM_OVER:
			write_byte(Addr,NewByte);
			break;
		case DM_AND:
			write_byte(Addr,OldByte & NewByte);
			break;
		case DM_OR:
			write_byte(Addr,OldByte | NewByte);
			break;			
		case DM_XOR:
			write_byte(Addr,OldByte ^ NewByte);
			break;
	    }

}


void 
DrawLineHorz(uns8 x, uns8 y, uns8 Length, uns8 Colour, int DrawMode)
{

	uns16 ScrnByte = BytePos(x,y);
	int LengthLeft = Length;
	
	if (RightNybble(x)) {
		PlotRight(ScrnByte++, Colour, DrawMode);		
		LengthLeft--;
	}

	while (LengthLeft > 1) {
		PlotBoth(ScrnByte++,((Colour & R_MASK) << 4 | (Colour & R_MASK)), DrawMode);
		LengthLeft -=2;
	}

	if (LengthLeft !=0 ) {
		PlotLeft(ScrnByte, Colour, DrawMode);
	}

}

void 
DrawLineVert(uns8 x, uns8 y, uns8 Length, uns8 Colour, int DrawMode)
{

	uns16 ScrnByte = BytePos(x,y);
	int LengthLeft = Length;	

	for (; LengthLeft--; ScrnByte+=128) {
	
		if (RightNybble(x))
			PlotRight(ScrnByte, Colour, DrawMode);
		else
			PlotLeft(ScrnByte, Colour, DrawMode); 
	}
}

void
FillArea(uns8 x, uns8 y, uns8 width, uns8 height, uns8 Colour, int DrawMode)
{
	int i;
	
	for (i=0; i<height; i++)
		DrawLineHorz(x, y++, width, Colour, DrawMode);

}

void
DrawChar(uns8 x, uns8 y, char Character, uns8 Colour, int DrawMode)
{
	int i,j;
	uns16 ScrnByte = BytePos(x,y);
	uns8 *GlyphPtr = DialogFontPtr + FontOffset(Character)+1;
	
	/* Fill both Nybbles with Colour */

	Colour = ((Colour & R_MASK) << 4 | (Colour & R_MASK));

	for (i = 0; i < GLYPH_HEIGHT; i++) {
#if 0
		PlotBoth(ScrnByte++, Colour & *(GlyphPtr  ), DrawMode);
		PlotBoth(ScrnByte++, Colour & *(GlyphPtr+1), DrawMode);
		PlotBoth(ScrnByte++, Colour & *(GlyphPtr+2), DrawMode);
		GlyphPtr+=3;
#else
		for (j = 0; j < 3; j++) {
			switch (*GlyphPtr++) {
				case 0x77:
					PlotBoth (ScrnByte++, Colour, DrawMode);
					break;
				case 0x70:
					PlotLeft (ScrnByte++, Colour, DrawMode);
					break;
				case 0x07:
					PlotRight(ScrnByte++, Colour, DrawMode);
					break;
				case 0x00:
					ScrnByte++;	
					break;
			}

		}
#endif
		ScrnByte+=125;
	}
}

void
DrawString(uns8 x, uns8 y, char *TheString, uns8 Colour, int DrawMode)
{

	int  i;
	
	for (i = 0; i < strlen(TheString); i++) {
	
		DrawChar(x, y, TheString[i], Colour, DrawMode);
		x+=GLYPH_WIDTH;
	}

}

void
DrawShadowBox(uns8 x, uns8 y, uns8 width, uns8 height, int ShadowMode, int FillBox, int DrawMode)
{

	uns8 TLColour, BRColour;

	/* Sort out the Colours to use */

	switch(ShadowMode) {
	
		case SHADOW_OUT:
			TLColour = SHCOL_LIGHT;
			BRColour = SHCOL_DARK;
			break;
		case SHADOW_IN:
			TLColour = SHCOL_DARK;
			BRColour = SHCOL_LIGHT;
			break;
	}

	/* Now draw the borders of the box */
	
	DrawLineHorz(x, y, width,      TLColour, DrawMode);
	DrawLineHorz(x, y+1, width-1,  TLColour, DrawMode);
	DrawLineVert(x, y, height,     TLColour, DrawMode);
	DrawLineVert(x+1, y, height-1, TLColour, DrawMode);	

	DrawLineHorz(x+1, y+height-1, width-1, BRColour, DrawMode);
	DrawLineHorz(x+2, y+height-2, width-2, BRColour, DrawMode);
	DrawLineVert(x+width-1, y+1, height-1, BRColour, DrawMode);
	DrawLineVert(x+width-2, y+2, height-2, BRColour, DrawMode);

	/* Fill the box if required */
	
	if (FillBox)
		FillArea(x+2,y+2,width-4, height-4, SHCOL_MEDIUM, DrawMode);

}

void
DrawBox(uns8 x, uns8 y, uns8 width, uns8 height, int Colour, int DrawMode)
{
	
	DrawLineHorz(x, y, width,      Colour, DrawMode);
	DrawLineHorz(x, y+1, width-1,  Colour, DrawMode);
	DrawLineVert(x, y, height,     Colour, DrawMode);
	DrawLineVert(x+1, y, height-1, Colour, DrawMode);	

	DrawLineHorz(x+1, y+height-1, width-1, Colour, DrawMode);
	DrawLineHorz(x+2, y+height-2, width-2, Colour, DrawMode);
	DrawLineVert(x+width-1, y+1, height-1, Colour, DrawMode);
	DrawLineVert(x+width-2, y+2, height-2, Colour, DrawMode);

}

void ClearDisplay(void)
{
	/* Clear the whole display */

	uns16 ScrnByte = BytePos(0,0);
	int i;
	
	for (i=0; i < 24576; i++)
		write_byte(ScrnByte++, 0x00);

}

void
FetchArea(uns8 *Buffer, uns8 x, uns8 y, uns8 width, uns8 height)
{
	/* Fetch the required display area and store in the (prealloced) */
	/* Memory pointed to by Buffer					 */

	uns16 ScrnByte = BytePos(x,y);
	int ByteWidth = BytePos(x+width+1,y) - ScrnByte;
	int i;
	
	while (height--) {
		for (i=0; i < ByteWidth; i++)
			*Buffer++ = read_byte(ScrnByte++);
		ScrnByte+= (128-ByteWidth);
	}
}



void
StoreArea(uns8 *Buffer, uns8 x, uns8 y, uns8 width, uns8 height)
{

	/* Store  the required display area from the (preallocated) */
	/* Memory pointed to by Buffer				    */

	uns16 ScrnByte = BytePos(x,y);
	int ByteWidth = BytePos(x+width+1,y) - ScrnByte;
	int i;
	
	while (height--) {
		for (i=0; i < ByteWidth; i++)
			if (IsOnScreen(ScrnByte))
				write_byte(ScrnByte++, *Buffer++);

		ScrnByte+= (128-ByteWidth);
	}
}

void
DrawPixmap(uns8 x, uns8 y, Pixmap *ThePixmap, int DrawMode)
{


	/* Draw ThePixmap at coordinates (x, y) */

	int height   = ThePixmap->PixmapHeight;
	uns8 *Buffer = ThePixmap->PixmapData;

	uns16 ScrnByte = BytePos(x,y);
	int ByteWidth = BytePos(x+ThePixmap->PixmapWidth+1,y) - ScrnByte;

	
	while (height--) {
	
		int LengthLeft = ThePixmap->PixmapWidth;
	
		if (RightNybble(x)) {
			if (x+(ThePixmap->PixmapWidth-LengthLeft) < DISPLAY_WIDTH)
				PlotRight(ScrnByte++, *Buffer++, DrawMode);		
			else {
				ScrnByte++;
				Buffer++;	
			}
			LengthLeft--;
		}

		while (LengthLeft > 1) {
			if (x+(ThePixmap->PixmapWidth-LengthLeft) < DISPLAY_WIDTH)
				PlotBoth(ScrnByte++,((*(Buffer+0) & R_MASK) << 4 | 
			                     (*(Buffer+1) & R_MASK)), DrawMode);
			else
				ScrnByte++;

			Buffer     +=2;
			LengthLeft -=2;
		}
	
		if (LengthLeft !=0 ) {
			if (x+(ThePixmap->PixmapWidth-LengthLeft) < DISPLAY_WIDTH)
				PlotLeft(ScrnByte++, *Buffer++, DrawMode);
			else {
				ScrnByte++;
				Buffer++;
			}
		}

		ScrnByte+= (128-ByteWidth);
	
	}
	
}
