/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File vgasamgrx.c : SAM display driver routines - Linux vga version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <vga.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"

extern uns8 *vga_data_ptr;
extern int flash_phase;
extern int ClutChangeInFrame;
extern int BorderChangeInFrame;
extern int VmprChangeInFrame;	
extern int Vga2x2LinearAddressing;

/* Used for paged vga addressing for 2x2 scale mode */

static int SwitchOverLines[4] = { 27, 78, 129, 180 };
static int VgaPageNo;
static uns8 TmpBuff[1280];
static uns8 *TmpPtr = NULL; 
static uns8 *vga_data_end;
static int VgaMidPoint;

/* These macros map a byte in the attribute area to an vga colour index */
#define Background(byte) (((byte) & 0x38) >> 3) | (((byte) & 0x40) >> 3)
#define Foreground(byte) ((byte) & 0x07) | (((byte) & 0x40) >> 3)

uns8 LineNeedsUpdating[256];	/* Determines whether a given line needs to
				   be redrawn */

static uns8 *SamMemoryPtr;	/* Pointer to the SAM screen memory */
static uns8 *VgaMemoryPtr;	/* Pointer within the vga frame buffer */

void (*borderline_update)(void);
void (*line_update)(int);
void (*line_update_mode_1)(int);
void (*line_update_mode_2)(int);
void (*line_update_mode_3)(int);
void (*line_update_mode_4)(int);

int BorderTop;
int BorderBottom;

void
borderline_update_1x1(void)
{
/* Update either top or bottom borders if required */

   if (ClutChangeInFrame || BorderChangeInFrame)
   	memset(VgaMemoryPtr, clut[BORD_VAL(border)], 320);
   
   VgaMemoryPtr += 320;
   	
}

void
line_update_vga_mode1_1x1(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8* SamAttrPtr;
   int i;

   extern unsigned int y_dtbl[];

/* In Mode 1 we have to update SamMemoryPtr each line due to the 
   spectrum screen layout, but we use a lookup table to minimize
   the cost */

   SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
   SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
   
/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++;
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}

		i_pix = clut[ink];
		p_pix = clut[paper];
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	VgaMemoryPtr += 320;		/* 320 Bytes in a vga 1x1 line */
   }
}

void
line_update_vga_mode2_1x1(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8* SamAttrPtr;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr once in a given frame, at LineNo == 0 */
   
   if (VmprChangeInFrame || !LineNo)
   	SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	SamAttrPtr = SamMemoryPtr + 0x2000;
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++; 
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}
		i_pix = clut[ink];
		p_pix = clut[paper];
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	VgaMemoryPtr += 320;		/* 320 Bytes in a vga 1x1 line */
   }
}


void
line_update_vga_mode3_1x1(int LineNo)
{
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(VgaMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x30) >> 4)];
		*(VgaMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x03))];
		SamMemoryPtr++;
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 3 line */
	VgaMemoryPtr += 320;		/* 320 Bytes in a vga 1x1 line */
   }
}

void
line_update_vga_mode4_1x1(int LineNo)
{
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
     BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	
	for (i=0; i < 128; i++) {
		*(VgaMemoryPtr++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(VgaMemoryPtr++) = clut[ (*SamMemoryPtr++) &0x0f];
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	VgaMemoryPtr += BorderWidth_H_1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 4 line */
	VgaMemoryPtr += 320;		/* 320 Bytes in a vga 1x1 line */
   }
}


void
borderline_update_2x2(void)
{
/* Update either top or bottom borders if required */

   if (ClutChangeInFrame || BorderChangeInFrame)
   	memset(VgaMemoryPtr, clut[BORD_VAL(border)], 1280);
   
   VgaMemoryPtr += 1280;
   	
}


void
line_update_vga_mode1_2x2(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8 *VgaMemoryPtrSave;
   uns8* SamAttrPtr;
   int i;

   extern unsigned int y_dtbl[];

/* In Mode 1 we have to update SamMemoryPtr each line due to the 
   spectrum screen layout, but we use a lookup table to minimize
   the cost */

   SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
   SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
   
/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* If we are using a paged vga mode for 2x2 scaling we need to deal with
	   the switch over lines in a special way */
   
	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		TmpPtr = VgaMemoryPtr;
		VgaMemoryPtr = TmpBuff;   
	}

  	/* Save the Vga pointer - we need this for generating the 
  	   next line down - using memcpy */

	VgaMemoryPtrSave = VgaMemoryPtr;

	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++;
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}

		i_pix = clut[ink];
		p_pix = clut[paper];
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x01) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;


	/* Now generate the 2nd line by copying the first */
	memcpy(VgaMemoryPtr, VgaMemoryPtrSave, 640);
	VgaMemoryPtr += 640;

	if (TmpPtr != NULL) {
		VgaMemoryPtr = TmpPtr;
		VgaMidPoint = (vga_data_end - VgaMemoryPtr);
		memcpy(VgaMemoryPtr, TmpBuff, VgaMidPoint); 
		vga_setpage(++VgaPageNo);
		VgaMemoryPtr = vga_data_ptr;
		memcpy(VgaMemoryPtr, &TmpBuff[VgaMidPoint], (1280-VgaMidPoint));
		VgaMemoryPtr += (1280-VgaMidPoint);
		TmpPtr = NULL;
	}

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */

	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		VgaMidPoint = (vga_data_end-VgaMemoryPtr);
		VgaMemoryPtr = vga_data_ptr + (1280-VgaMidPoint);
		vga_setpage(++VgaPageNo);
	}
	else {
		VgaMemoryPtr += 1280;		/* 1280 Bytes in a vga 2x2 line */
 	}
   }
}
void
line_update_vga_mode2_2x2(int LineNo)
{
   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8 *VgaMemoryPtrSave;
   uns8* SamAttrPtr;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr once in a given frame, at LineNo == 0 */
   
   if (VmprChangeInFrame || !LineNo)
   	SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   

	/* If we are using a paged vga mode for 2x2 scaling we need to deal with
	   the switch over lines in a special way */
   
	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		TmpPtr = VgaMemoryPtr;
		VgaMemoryPtr = TmpBuff;   
	}

  	/* Save the Vga pointer - we need this for generating the 
  	   next line down - using memcpy */

	VgaMemoryPtrSave = VgaMemoryPtr;
   
	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	SamAttrPtr = SamMemoryPtr + 0x2000;
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++; 
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}
		i_pix = clut[ink];
		p_pix = clut[paper];
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*SamMemoryPtr & 0x01) ? i_pix : p_pix;
		*(VgaMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;

	/* Now generate the 2nd line by copying the first */
	memcpy(VgaMemoryPtr, VgaMemoryPtrSave, 640);
	VgaMemoryPtr += 640;

	if (TmpPtr != NULL) {
		VgaMemoryPtr = TmpPtr;
		VgaMidPoint = (vga_data_end - VgaMemoryPtr);
		memcpy(VgaMemoryPtr, TmpBuff, VgaMidPoint); 
		vga_setpage(++VgaPageNo);
		VgaMemoryPtr = vga_data_ptr;
		memcpy(VgaMemoryPtr, &TmpBuff[VgaMidPoint], (1280-VgaMidPoint));
		VgaMemoryPtr += (1280-VgaMidPoint);
		TmpPtr = NULL;
	}

	LineNeedsUpdating[LineNo] = False;


   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */

	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		VgaMidPoint = (vga_data_end-VgaMemoryPtr);
		VgaMemoryPtr = vga_data_ptr + (1280-VgaMidPoint);
		vga_setpage(++VgaPageNo);
	}
	else {
		VgaMemoryPtr += 1280;		/* 1280 Bytes in a vga 2x2 line */
 	}
   }
}
void
line_update_vga_mode3_2x2(int LineNo)
{
   uns8 *VgaMemoryPtrSave;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* If we are using a paged vga mode for 2x2 scaling we need to deal with
	   the switch over lines in a special way */
   
	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		TmpPtr = VgaMemoryPtr;
		VgaMemoryPtr = TmpBuff;   
	}


  	/* Save the Vga pointer - we need this for generating the 
  	   next line down - using memcpy */

	VgaMemoryPtrSave = VgaMemoryPtr;

	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(VgaMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0xc0) >> 6)];
		*(VgaMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x30) >> 4)];
		*(VgaMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x0c) >> 2)];
		*(VgaMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x03))];
		SamMemoryPtr++;
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;

	/* Now generate the 2nd line by copying the first */
	memcpy(VgaMemoryPtr, VgaMemoryPtrSave, 640);
	VgaMemoryPtr += 640;

	if (TmpPtr != NULL) {
		VgaMemoryPtr = TmpPtr;
		VgaMidPoint = (vga_data_end - VgaMemoryPtr);
		memcpy(VgaMemoryPtr, TmpBuff, VgaMidPoint); 
		vga_setpage(++VgaPageNo);
		VgaMemoryPtr = vga_data_ptr;
		memcpy(VgaMemoryPtr, &TmpBuff[VgaMidPoint], (1280-VgaMidPoint));
		VgaMemoryPtr += (1280-VgaMidPoint);
		TmpPtr = NULL;
	}

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 3 line */

	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		VgaMidPoint = (vga_data_end-VgaMemoryPtr);
		VgaMemoryPtr = vga_data_ptr + (1280-VgaMidPoint);
		vga_setpage(++VgaPageNo);
	}
	else {
		VgaMemoryPtr += 1280;		/* 1280 Bytes in a vga 2x2 line */
 	}
   }
}

void
line_update_vga_mode4_2x2(int LineNo)
{
   uns8 *VgaMemoryPtrSave;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }



/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
     BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   

	/* If we are using a paged vga mode for 2x2 scaling we need to deal with
	   the switch over lines in a special way */
   
	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		TmpPtr = VgaMemoryPtr;
		VgaMemoryPtr = TmpBuff;   
	}

  	/* Save the Vga pointer - we need this for generating the 
  	   next line down - using memcpy */

	VgaMemoryPtrSave = VgaMemoryPtr;
   
	/* Lhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(VgaMemoryPtr++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(VgaMemoryPtr++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(VgaMemoryPtr++) = clut[ (*SamMemoryPtr)   &0x0f];
		*(VgaMemoryPtr++) = clut[ (*SamMemoryPtr++) &0x0f];
	}
	
	/* Rhs Border */
	memset(VgaMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	VgaMemoryPtr += BorderWidth_H_2x2;


	/* Now generate the 2nd line by copying the first */
	memcpy(VgaMemoryPtr, VgaMemoryPtrSave, 640);
	VgaMemoryPtr += 640;

	if (TmpPtr != NULL) {
		VgaMemoryPtr = TmpPtr;
		VgaMidPoint = (vga_data_end - VgaMemoryPtr);
		memcpy(VgaMemoryPtr, TmpBuff, VgaMidPoint); 
		vga_setpage(++VgaPageNo);
		VgaMemoryPtr = vga_data_ptr;
		memcpy(VgaMemoryPtr, &TmpBuff[VgaMidPoint], (1280-VgaMidPoint));
		VgaMemoryPtr += (1280-VgaMidPoint);
		TmpPtr = NULL;
	}

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 4 line */

	if (!Vga2x2LinearAddressing && (LineNo == SwitchOverLines[VgaPageNo])) {
		VgaMidPoint = (vga_data_end-VgaMemoryPtr);
		VgaMemoryPtr = vga_data_ptr + (1280-VgaMidPoint);
		vga_setpage(++VgaPageNo);
	}
	else {
		VgaMemoryPtr += 1280;		/* 1280 Bytes in a vga 2x2 line */
 	}
   }
}

void
FlyBack(void)
{
/* Reset the vga frame buffer pointer */

   VgaMemoryPtr = vga_data_ptr;

/* Reset vga paging variables if required */

   if (!Vga2x2LinearAddressing){
	VgaPageNo = 0;   	
	vga_setpage(0);
   }   
}


void
ChangeScaleMode(int SF)
{

	int i;
	
	if (SF == 1) {
	
		borderline_update      = borderline_update_1x1;
		line_update_mode_1 = line_update_vga_mode1_1x1;
		line_update_mode_2 = line_update_vga_mode2_1x1;
		line_update_mode_3 = line_update_vga_mode3_1x1;
		line_update_mode_4 = line_update_vga_mode4_1x1;
		vga_setmode(VGA1X1MODE);
		BorderTop    = 55; 
		BorderBottom = 255;
	}
	else {
	
		borderline_update      = borderline_update_2x2;
		line_update_mode_1 = line_update_vga_mode1_2x2;
		line_update_mode_2 = line_update_vga_mode2_2x2;
		line_update_mode_3 = line_update_vga_mode3_2x2;
		line_update_mode_4 = line_update_vga_mode4_2x2;
		vga_setmode(VGA2X2MODE);	

		BorderTop    = 35; 
		BorderBottom = 286;

		if (Vga2x2LinearAddressing)
			  vga_setlinearaddressing();

	}
	vga_data_ptr = vga_getgraphmem();
	vga_data_end = vga_data_ptr + 0x10000;
	get_colours();
	/* Make sure we update the whole screen */
	for (i=0; i < 192; i++)
		LineNeedsUpdating[i] = True;
	BorderChangeInFrame = (192+119);
	FlyBack();
}
